import os
from typing import Tuple

import torch
from fastchat.model import load_model

try:
    from vllm import LLM as VLLM
    from vllm import SamplingParams
except:
    LLM = None
    SamplingParams = None
from transformers.generation import GenerationConfig

from ragnarok.data import RAGExecInfo, Request
from ragnarok.generate.llm import LLM, PromptMode
from ragnarok.generate.post_processor import GPTPostProcessor
from ragnarok.generate.templates.chat_qa import ChatQATemplate


class OSLLM(LLM):
    def __init__(
        self,
        model: str,
        context_size: int = 8192,
        prompt_mode: PromptMode = PromptMode.CHATQA,
        max_output_tokens: int = 1500,
        num_few_shot_examples: int = 0,
        device: str = "cuda",
        num_gpus: int = 1,
        dtype: str = "float16",
    ) -> None:
        """
        Creates instance of the OSLLM class, an extension of LLM designed for performing retrieval-augmented generation using
        a specified open-source language model. Some configurations are supported such as GPU acceleration, and custom system
        messages for generating prompts.

        Parameters:
        - model (str): Identifier for the language model to be used for ranking tasks.
        - context_size (int, optional): Maximum number of tokens that can be handled in a single prompt. Defaults to 4096.
        - prompt_mode (PromptMode, optional): Specifies the mode of prompt generation, with the default set to RAGNAROK,
        indicating that this class is designed primarily for listwise ranking tasks following the RAGNAROK methodology.
        - max_output_tokens (int, optional): Maximum number of tokens that can be generated in a single response. Defaults to 1500.
        - num_few_shot_examples (int, optional): Number of few-shot learning examples to include in the prompt, allowing for
        the integration of example-based learning to improve model performance. Defaults to 0, indicating no few-shot examples
        by default.
        - device (str, optional): Specifies the device for model computation ('cuda' for GPU or 'cpu'). Defaults to 'cuda'.
        - num_gpus (int, optional): Number of GPUs to use for model loading and inference. Defaults to 1.
        - dtype (str, optional): Specifies the data type for model computation ('float32' or 'float16' or 'bfloat16'). Defaults to 'float32'.
        Raises:
        - AssertionError: If CUDA is specified as the device but is not available on the system.
        - ValueError: If an unsupported prompt mode is provided.

        Note:
        - This class is operates given scenarios where listwise ranking is required, with support for dynamic
        passage handling and customization of prompts through system messages and few-shot examples.
        - GPU acceleration is supported and recommended for faster computations.
        """
        super().__init__(
            model, context_size, prompt_mode, max_output_tokens, num_few_shot_examples
        )
        self._device = device
        if self._device == "cuda":
            assert torch.cuda.is_available()
        if prompt_mode not in [PromptMode.CHATQA]:
            raise ValueError(
                f"Unsupported prompt mode: {prompt_mode}. The only prompt mode currently supported is a slight variation of {PromptMode.CHATQA} prompt."
            )
        # ToDo: Make repetition_penalty configurable
        try:
            print(model)
            self._llm = VLLM(
                model,
                download_dir=os.getenv("HF_HOME"),
                enforce_eager=False,
                tensor_parallel_size=2,
            )
            self._tokenizer = self._llm.get_tokenizer()
        except:
            self._llm, self._tokenizer = load_model(
                model, device=device, num_gpus=num_gpus
            )
        self._post_processor = GPTPostProcessor()
        if num_few_shot_examples > 0:
            # TODO(ronak): Add support for few-shot examples
            pass

    def run_llm(
        self, prompt: str, logging: bool = False, vllm: bool = True
    ) -> Tuple[str, int]:
        if logging:
            print(f"Prompt: {prompt}")
        try:
            inputs = self._tokenizer([prompt])
            sampling_params = SamplingParams(
                temperature=0.0,
                max_tokens=self._output_token_estimate,
                min_tokens=100,
            )
            output = self._llm.generate([prompt], sampling_params)
            text = output[0].outputs[0].text
            # Anything after a User: / Context: / References: / Note: remove
            text = (
                text.split("User:")[0]
                .split("Context:")[0]
                .split("References:")[0]
                .split("Note:")[0]
            )
            if logging:
                print(f"Response: {text}")
            answers, rag_exec_response = self._post_processor(text)
            if logging:
                print(f"Answer: {answers}")
            rag_exec_info = RAGExecInfo(
                prompt=prompt,
                response=rag_exec_response,
                input_token_count=self.get_num_tokens(prompt),
                output_token_count=sum([len(ans.text) for ans in answers]),
                candidates=[],
            )
            return answers, rag_exec_info
        except:
            inputs = {k: torch.tensor(v).to(self._device) for k, v in inputs.items()}
            gen_cfg = GenerationConfig.from_model_config(self._llm.config)
            gen_cfg.max_new_tokens = self.num_output_tokens()
            # gen_cfg.temperature = 0
            gen_cfg.do_sample = False
            output_ids = self._llm.generate(**inputs, generation_config=gen_cfg)

            if self._llm.config.is_encoder_decoder:
                output_ids = output_ids[0]
            else:
                output_ids = output_ids[0][len(inputs["input_ids"][0]) :]
            outputs = self._tokenizer.decode(
                output_ids,
                skip_special_tokens=True,
                spaces_between_special_tokens=False,
            )
            return outputs, output_ids.size(0)

    def create_prompt(self, request: Request, topk: int) -> Tuple[str, int]:
        query = request.query.text
        max_length = (self._context_size - 200) // topk
        while True:
            rank = 0
            context = []
            for cand in request.candidates[:topk]:
                rank += 1
                content = self.convert_doc_to_prompt_content(cand.doc, max_length)
                context.append(
                    f"[{rank}] {self._replace_number(content)}",
                )
            if self._prompt_mode == PromptMode.CHATQA:
                chat_qa_template = ChatQATemplate()
                messages = chat_qa_template(query, context, "chatqa")
            else:
                raise ValueError(
                    f"Unsupported prompt mode: {self._prompt_mode}, expected {PromptMode.CHATQA}."
                )
            num_tokens = self.get_num_tokens(messages)
            if num_tokens <= self.max_tokens() - self.num_output_tokens():
                break
            else:
                max_length -= max(
                    1,
                    (num_tokens - self.max_tokens() + self.num_output_tokens())
                    // (topk * 4),
                )
        return messages, self.get_num_tokens(messages)

    def get_num_tokens(self, prompt: str) -> int:
        return len(self._tokenizer.encode(prompt))

    def cost_per_1k_token(self, input_token: bool) -> float:
        return 0



from typing import Any, Dict, List, Union

from ragnarok.data import Query, Request

# from ragnarok.evaluation.nugget_eval import EvalFunction
from ragnarok.generate.api_keys import get_azure_openai_args, get_openai_api_key
from ragnarok.generate.cohere import Cohere
from ragnarok.generate.generator import RAG
from ragnarok.generate.gpt import SafeOpenai
from ragnarok.generate.llm import PromptMode
from ragnarok.generate.os_llm import OSLLM
from ragnarok.retrieve_and_rerank.restriever import Restriever
from ragnarok.retrieve_and_rerank.retriever import (
    CacheInputFormat,
    RetrievalMethod,
    RetrievalMode,
    Retriever,
)


def retrieve_and_generate(
    generator_path: str,
    dataset: Union[str, List[str], List[Dict[str, Any]]],
    retrieval_mode: RetrievalMode = RetrievalMode.DATASET,
    retrieval_method: List[RetrievalMethod] = [
        RetrievalMethod.BM25,
        RetrievalMethod.RANK_ZEPHYR_RHO,
    ],
    k: List[int] = [100, 20],
    context_size: int = 8192,
    max_output_tokens: int = 1500,
    device: str = "cuda",
    num_gpus: int = 2,
    prompt_mode: PromptMode = PromptMode.CHATQA,
    num_few_shot_examples: int = 0,
    shuffle_candidates: bool = False,
    print_prompts_responses: bool = False,
    query: str = "",
    qid: int = 1,
    use_azure_openai: bool = False,
    num_passes: int = 1,
    window_size: int = 20,
    step_size: int = 10,
    interactive: bool = False,
    host_reranker: str = "8082",
    host_retriever: str = "8081",
):
    """orchestrates 3 stage RAG process: Retrieval (e.g., BM25), reranking (e.g., RankZephyr), generation (e.g., GPT-4o)

    Args:
        generator_path (str): model name for generator
        dataset (str): dataset from which to search from
        retrieval_mode (RetrievalMode): mode of retrieval (e.g., DATASET)
        k (List[int]): [top_k_retrieve, top_k_rerank]. The top top_k_retrieve elements to retrieve from the dataset then the top top_k_rerank elements to return after reranking.
        qid (int): QID of the search query
        query (str): The query to search for
        interactive (bool): Setting interactive to true will call the Reranker API. Otherwise will obtain from pre-cached data
        host_reranker (str): Host name of the Reranker API (will call Retriever API, so we need to pass in the retriever host)
        host_retriever (str): Host name of the Retriever API

    Return:
        Returns the generation results in JSON format specified by TREC 2024
    """

    # Construct Generation Agent
    model_full_path = ""
    if "gpt" in generator_path:
        print("Using OpenAI API")
        print(generator_path)
        print("Using Azure OpenAI API" if use_azure_openai else "")
        openai_keys = get_openai_api_key()
        agent = SafeOpenai(
            model=generator_path,
            context_size=context_size,
            prompt_mode=prompt_mode,
            max_output_tokens=max_output_tokens,
            num_few_shot_examples=num_few_shot_examples,
            keys=openai_keys,
            **(get_azure_openai_args() if use_azure_openai else {}),
        )
    elif "command-r" in generator_path:
        agent = Cohere(
            model=generator_path,
            context_size=context_size,
            prompt_mode=prompt_mode,
            max_output_tokens=max_output_tokens,
            num_few_shot_examples=num_few_shot_examples,
        )
    elif "llama" in generator_path.lower() or "mistral" in generator_path.lower():
        agent = OSLLM(
            model=generator_path,
            context_size=context_size,
            prompt_mode=prompt_mode,
            max_output_tokens=max_output_tokens,
            num_few_shot_examples=num_few_shot_examples,
            device=device,
            num_gpus=num_gpus,
        )
    else:
        raise ValueError(f"Unsupported model: {generator_path}")

    # Retrieve + Rerank
    print("Calling reranker API...")
    # Only DATASET mode is currently supported.
    if retrieval_mode == RetrievalMode.DATASET:
        if interactive:
            # Calls the host_reranker API to obtain the results after first 2 stages (retrieve+rerank)
            requests = [
                Restriever.from_dataset_with_prebuilt_index(
                    dataset_name=dataset,
                    retrieval_method=retrieval_method,
                    host_reranker=host_reranker,
                    host_retriever=host_retriever,
                    request=Request(query=Query(text=query, qid=qid)),
                    k=k,
                )
            ]
        else:
            requests = Retriever.from_dataset_with_prebuilt_index(
                dataset_name=dataset,
                retrieval_method=retrieval_method,
                k=k,
                cache_input_format=CacheInputFormat.JSONL,
            )
            print()
    else:
        raise ValueError(f"Invalid retrieval mode: {retrieval_mode}")

    # Generation
    print("Generating...")
    rag = RAG(agent)
    rag_results = rag.answer_batch(
        requests,
        topk=k[-1],
        shuffle_candidates=shuffle_candidates,
        logging=print_prompts_responses,
    )
    if isinstance(dataset, str):
        file_name = rag.write_answer_results(
            retrieval_method[-1].name,
            rag_results,
            shuffle_candidates,
            top_k_candidates=k[-1],
            dataset_name=dataset,
        )
        print(f"Results written to {file_name}")

    return rag_results[0]



import argparse
import os
import sys

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
parent = os.path.dirname(SCRIPT_DIR)
parent = os.path.dirname(parent)
sys.path.append(parent)

import torch

from ragnarok.generate.llm import PromptMode
from ragnarok.retrieve_and_generate import retrieve_and_generate
from ragnarok.retrieve_and_rerank.retriever import RetrievalMethod, RetrievalMode
from ragnarok.retrieve_and_rerank.topics_dict import TOPICS


def parse_topk(value):
    try:
        return [int(k) for k in value.split(",")]
    except ValueError:
        raise argparse.ArgumentTypeError(
            f"Invalid comma-separated list of integers: {value}"
        )


def parse_retrieval_methods(value):
    try:
        # Ensure it is of type RetrievalMethod
        return [RetrievalMethod(e) for e in value.split(",")]
    except ValueError:
        raise argparse.ArgumentTypeError(
            f"Invalid comma-separated list of retrieval methods: {value}"
        )


def main(args):
    model_path = args.model_path
    use_azure_openai = args.use_azure_openai
    context_size = args.context_size
    topk = args.topk
    dataset = args.dataset
    num_gpus = args.num_gpus
    retrieval_method = args.retrieval_method
    prompt_mode = args.prompt_mode
    num_few_shot_examples = args.num_few_shot_examples
    shuffle_candidates = args.shuffle_candidates
    print_prompts_responses = args.print_prompts_responses
    num_few_shot_examples = args.num_few_shot_examples
    max_output_tokens = args.max_output_tokens
    device = "cuda" if torch.cuda.is_available() else "cpu"
    retrieval_mode = RetrievalMode.DATASET

    _ = retrieve_and_generate(
        model_path,
        dataset,
        retrieval_mode,
        retrieval_method,
        topk,
        context_size,
        max_output_tokens,
        device,
        num_gpus,
        prompt_mode,
        num_few_shot_examples,
        shuffle_candidates,
        print_prompts_responses,
        use_azure_openai=use_azure_openai,
    )


""" sample run:
python src/ragnarok/scripts/run_ragnarok.py  --model_path=cohere_command_r_plus  --topk=20 --dataset=researchy-questions  --retrieval_method=bm25,rank_zephyr --prompt_mode=chatqa  --context_size=8192 --max_output_tokens=1500
"""
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_path",
        type=str,
        required=True,
        help="Path to the model. If `use_azure_ai`, pass your deployment name.",
    )
    parser.add_argument(
        "--use_azure_openai",
        action="store_true",
        help="If True, use Azure OpenAI instead of regular OpenAI. Requires env var to be set: "
        "`AZURE_OPENAI_API_VERSION`, `AZURE_OPENAI_API_BASE`",
    )
    parser.add_argument(
        "--context_size", type=int, default=8192, help="context size used for model"
    )
    parser.add_argument(
        "--topk",
        type=parse_topk,
        default=[100, 20],
        help="Comma-separated list of top k values for each retrieval method. Example: 100,20",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        required=True,
        help=f"Should be one of 1- dataset name, must be in {TOPICS.keys()},  2- a list of inline documents  3- a list of inline hits 4- filename containing retrieved results",
    )
    parser.add_argument(
        "--num_gpus", type=int, default=1, help="the number of GPUs to use"
    )
    parser.add_argument(
        "--retrieval_method",
        type=parse_retrieval_methods,
        required=True,
        help="Comma-separated list of retrieval methods. Choices: "
        + ", ".join([e.value for e in RetrievalMethod]),
    )
    parser.add_argument(
        "--prompt_mode",
        type=PromptMode,
        required=True,
        choices=list(PromptMode),
    )
    parser.add_argument(
        "--shuffle_candidates",
        action="store_true",
        help="whether to shuffle the candidates before reranking",
    )
    parser.add_argument(
        "--print_prompts_responses",
        action="store_true",
        help="whether to print promps and responses",
    )
    parser.add_argument(
        "--num_few_shot_examples",
        type=int,
        required=False,
        default=0,
        help="number of in context examples to provide",
    )
    parser.add_argument(
        "--max_output_tokens",
        type=int,
        required=False,
        default=1500,
        help="maximum number of tokens in the output",
    )
    args = parser.parse_args()
    main(args)



from typing import List
from urllib import parse

import requests

from ragnarok.data import Candidate, Query, Request
from ragnarok.retrieve_and_rerank.retriever import RetrievalMethod, RetrievalMode


class Restriever:
    def __init__(
        self,
        retrieval_mode: RetrievalMode = RetrievalMode.DATASET,
        retrieval_method: List[RetrievalMethod] = [
            RetrievalMethod.BM25,
            RetrievalMethod.RANK_ZEPHYR,
        ],
    ) -> None:
        """
        Creates a Restriever instance with a specified retrieval method and mode.

        Args:
            retrieval_mode (RetrievalMode): The retrieval mode to be used. Defaults to DATASET. Only DATASET mode is currently supported.
            retrieval_method (List[RetrievalMethod]): A list of length 2 that contains: [retrieval method, reranking method]. Defaults to BM25 followed by RankZephyr.

        Raises:
            ValueError: If retrieval mode or retrieval method is invalid or missing.
        """
        self._retrieval_mode = retrieval_mode
        self._retrieval_method = retrieval_method

        if retrieval_mode != RetrievalMode.DATASET:
            raise ValueError(
                f"{retrieval_mode} is not supported for ServiceRetriever. Only DATASET mode is currently supported."
            )
        if not retrieval_method:
            raise "Please provide a retrieval method."
        if retrieval_method == RetrievalMethod.UNSPECIFIED:
            raise ValueError(
                f"Invalid retrieval method: {retrieval_method}. Please provide a specific retrieval method."
            )

    @staticmethod
    def from_dataset_with_prebuilt_index(
        dataset_name: str,
        host_reranker: str,
        host_retriever: str,
        retrieval_method: List[RetrievalMethod] = [
            RetrievalMethod.BM25,
            RetrievalMethod.RANK_ZEPHYR,
        ],
        k: List[int] = [100, 100],
        request: Request = None,
        num_passes: int = 1,
    ):
        """
        Creates a Restriever instance for a dataset with a prebuilt index.

        Args:
            - dataset_name (str): The name of the dataset.
            retrieval_method (Union[RetrievalMethod, List[RetrievalMethod]]): The retrieval method(s) to be used. Defaults to BM25.
            k (List[int], optional): The top k hits to retrieve. Defaults to [100, 20].

        Returns:
            List[Request]: The list of requests. Each request has a query and list of candidates.

        Raises:
            ValueError: If dataset name or retrieval method is invalid or missing.
        """
        if not dataset_name:
            raise ValueError("Please provide name of the dataset.")
        if not isinstance(dataset_name, str):
            raise ValueError(
                f"Invalid dataset format: {dataset_name}. Expected a string representing name of the dataset."
            )

        retriever = Restriever(
            RetrievalMode.DATASET,
            retrieval_method=retrieval_method,
        )
        return retriever.retrieve(
            dataset=dataset_name,
            k=k,
            host_retriever=host_retriever,
            host_reranker=host_reranker,
            request=request,
            num_passes=num_passes,
        )

    def retrieve(
        self,
        dataset: str,
        request: Request,
        host_reranker: str = "8082",
        host_retriever: str = "8081",
        k: List[int] = [20, 10],
        num_passes: int = 1,
    ) -> Request:
        """
        Executes the 2 stage retrieval+rerank process based on the configation provided with the Retriever instance. Takes in a Request object with a query and empty candidates object and the top k items to retrieve.

        Args:
            request (Request): The request containing the query and qid.
            dataset (str): The name of the dataset.
            k (List[int], optional): [top k hits to retrieve, top k hits to rerank]. Defaults to [20,10]
            host_reranker (str): The Anserini API host address. Defaults to 8081
            host_retriever (str): The Reranking API host address. Defaults to 8082
            num_passes (int): Number of passes to perform in the reranking stage

        Returns:
            Request. Contains a query and list of candidates
        Raises:
            ValueError: If the retrieval mode is invalid or the result format is not as expected.
        """

        parsed_query = parse.quote(request.query.text)
        (retrieval_method, rerank_method) = self._retrieval_method

        # API request URL specified in https://github.com/castorini/rank_llm api folder
        url = f"http://localhost:{host_reranker}/api/model/{rerank_method}/index/{dataset}/{host_retriever}?query={parsed_query}&hits_retriever={str(k[0])}&hits_reranker={str(k[1])}&qid={request.query.qid}&num_passes={num_passes}&retrieval_method={retrieval_method}"

        # Send request to Rerank API
        response = requests.get(url)

        # First 2 stages, retrieval and reranking, OK
        if response.ok:
            # Parse information about the query
            data = response.json()
            retrieved_results = Request(
                query=Query(text=data["query"]["text"], qid=data["query"]["qid"])
            )

            # Parse response into a list of candidates
            for candidate in data["candidates"]:
                retrieved_results.candidates.append(
                    Candidate(
                        docid=candidate["docid"],
                        score=candidate["score"],
                        doc=candidate["doc"],
                    )
                )

            print("Retrieval and reranking completed successfully...")
            print(
                f"Candidates provided during retrieval+reranking: {len(retrieved_results.candidates)}"
            )
        else:
            raise ValueError(
                f"Failed to retrieve data from RankLLM server. Error code: {response.status_code}"
            )
        return retrieved_results



import random
import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from ftfy import fix_text

from ragnarok.data import Request, Result, remove_unused_references


class PromptMode(Enum):
    UNSPECIFIED = "unspecified"
    COHERE = "cohere"
    CHATQA = "chatqa"

    def __str__(self):
        return self.value


class LLM(ABC):
    def __init__(
        self,
        model: str,
        context_size: int,
        prompt_mode: PromptMode,
        max_output_tokens: int = 1500,
        num_few_shot_examples: int = 0,
    ) -> None:
        self._model = model
        self._context_size = context_size
        self._prompt_mode = prompt_mode
        self._num_few_shot_examples = num_few_shot_examples
        self._output_token_estimate = max_output_tokens

    def max_tokens(self) -> int:
        """
        Returns the maximum number of tokens for a given model

        Returns:
            int: The maximum token count.
        """
        return self._context_size

    @abstractmethod
    def run_llm(
        self, prompt: Union[str, List[Dict[str, Any]]], logging: bool = False
    ) -> Tuple[Any, int]:
        """
        Abstract method to run the target language model with a passed in prompt.

        Args:
            prompt (Union[str, List[Dict[str, str]]]): The prompt to be processed by the model.
            logging (bool, optional): Flag to enable logging of operations. Defaults to False.

        Returns:
            Tuple[Any, int]: A tuple object containing the answer response and the number of tokens in the response.
        """
        pass

    @abstractmethod
    def create_prompt(
        self, request: Request, topk: int
    ) -> Tuple[Union[str, List[Dict[str, str]]], int]:
        """
        Abstract method to create a prompt based on the request and given topk.

        Args:
            request (Result): The request object containing data for prompt generation.
            topk (int): The topk ranks considered for prompt generation.

        Returns:
            Tuple[Union[str, List[Dict[str, str]]], int]: A tuple object containing the generated prompt and the number of tokens in the generated prompt.
        """
        pass

    @abstractmethod
    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """
        Abstract method to calculate the number of tokens contained in the given prompt.

        Args:
            prompt (Union[str, List[Dict[str, str]]]): The prompt for which to compute the token count for.

        Returns:
            int: The number of tokens in the given prompt.
        """
        pass

    @abstractmethod
    def cost_per_1k_token(self, input_token: bool) -> float:
        """
        Abstract method to calculate the cost per 1,000 tokens for the target language model.

        Args:
            input_token (bool): Flag to indicate if the cost is for input tokens or output tokens.

        Returns:
            float: The cost per 1,000 tokens.
        """
        pass

    @abstractmethod
    def num_output_tokens(self) -> int:
        """
        Abstract method to estimate the number of tokens in the model's output, constrained by max tokens for the target language model.

        Returns:
            int: The estimated number of output tokens.
        """
        pass

    def answer(
        self,
        request: Request,
        topk: int,
        shuffle_candidates: bool = False,
        logging: bool = False,
    ) -> Result:
        """
        Answer a given request using the target language model.

        Args:
            request (Request): The request object to process.
            topk (int): The topk ranks to consider for the generation process.
            shuffle_candidates (bool, optional): Flag to shuffle candidates before processing. Defaults to False.
            logging (bool, optional): Flag to enable logging of operations. Defaults to False.

        Returns:
            Result: The result object after answering the request.
        """
        return self.answer_batch(
            [request],
            topk,
            shuffle_candidates=shuffle_candidates,
            logging=logging,
        )[0]

    def answer_batch(
        self,
        requests: List[Request],
        topk: int,
        shuffle_candidates: bool = False,
        logging: bool = False,
    ) -> List[Result]:
        """
        Answer a list of requests using the target language model.

        Args:
            requests (List[Request]): The list of requests to process.
            topk (int): The topk ranks to consider for the generation process.
            shuffle_candidates (bool, optional): Flag to shuffle candidates before processing. Defaults to False.
            logging (bool, optional): Flag to enable logging of operations. Defaults to False.

        Returns:
            List[Result]: The list of results after answering the requests.
        """
        results = []
        for request in requests:
            if shuffle_candidates:
                # First randomly shuffle rerank_result in first topk ranks
                request.candidates[:topk] = random.sample(
                    request.candidates[:topk],
                    len(request.candidates[:topk]),
                )
            prompt, input_token_count = self.create_prompt(request, topk)
            answer, rag_exec_summary = self.run_llm(prompt, logging)
            rag_exec_summary.candidates = [
                candidate.__dict__ for candidate in request.candidates[:topk]
            ]
            result = Result(
                query=request.query,
                references=[cand.docid for cand in request.candidates[:topk]],
                answer=answer,
                rag_exec_summary=rag_exec_summary,
            )
            cleaned_result = remove_unused_references(result)
            if logging:
                print(f"Cleaned Result: {cleaned_result}")
            results.append(cleaned_result)
        return results

    def num_output_tokens(self) -> int:
        return self._output_token_estimate

    def _clean_response(self, response: str) -> str:
        new_response = fix_text(response)
        return new_response

    def _replace_number(self, s: str) -> str:
        return re.sub(r"\[(\d+)\]", r"(\1)", s)

    def convert_doc_to_prompt_content(
        self, doc: Dict[str, Any], max_length: int
    ) -> str:
        if "text" in doc:
            content = doc["text"]
        elif "segment" in doc:
            content = doc["segment"]
        elif "contents" in doc:
            content = doc["contents"]
        else:
            content = doc["passage"]
        if "title" in doc and doc["title"]:
            content = "Title: " + doc["title"] + " " + "Content: " + content
        content = content.strip()
        content = fix_text(content)
        content = content.replace("\n", " ")
        content = " ".join(content.split()[: int(max_length)])
        return self._replace_number(content)



import os
import sys
from pathlib import Path

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
parent = os.path.dirname(SCRIPT_DIR)
parent = os.path.dirname(parent)
sys.path.append(parent)

from ragnarok.data import DataWriter, read_requests_from_file
from ragnarok.generate.os_llm import OSLLM

file_name = "retrieve_results/BM25/retrieve_results_dl23_top20.json"
requests = read_requests_from_file(file_name)

generator = OSLLM("meta-llama/Meta-Llama-3-8B-Instruct")
rag_results = OSLLM.answer_batch(requests)
print(rag_results)

# write rerank results
writer = DataWriter(rag_results)
Path(f"demo_outputs/").mkdir(parents=True, exist_ok=True)
writer.write_in_json_format(f"demo_outputs/rag_results.json")
writer.write_in_trec_eval_format(f"demo_outputs/rerank_results.txt")
writer.write_ranking_exec_summary(f"demo_outputs/ranking_execution_summary.json")



import time
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

import openai
import tiktoken

from ragnarok.data import RAGExecInfo, Request
from ragnarok.generate.llm import LLM, PromptMode
from ragnarok.generate.post_processor import GPTPostProcessor
from ragnarok.generate.templates.chat_qa import ChatQATemplate


class SafeOpenai(LLM):
    def __init__(
        self,
        model: str,
        context_size: int,
        prompt_mode: PromptMode = PromptMode.CHATQA,
        max_output_tokens: int = 1500,
        num_few_shot_examples: int = 0,
        keys=None,
        key_start_id=None,
        proxy=None,
        api_type: str = None,
        api_base: str = None,
        api_version: str = None,
    ) -> None:
        """
        Creates instance of the SafeOpenai class, a specialized version of RankLLM designed for safely handling OpenAI API calls with
        support for key cycling, proxy configuration, and Azure AI conditional integration.

        Parameters:
        - model (str): The model identifier for the LLM (model identifier information can be found via OpenAI's model lists).
        - context_size (int): The maximum number of tokens that the model can handle in a single request.
        - prompt_mode (PromptMode, optional): Specifies the mode of prompt generation, with the default set to RANK_GPT,
         indicating that this class is designed primarily for listwise ranking tasks following the RANK_GPT methodology.
        - max_output_tokens (int, optional): Maximum number of tokens that can be generated in a single response. Defaults to 1500.
        - num_few_shot_examples (int, optional): Number of few-shot learning examples to include in the prompt, allowing for
        the integration of example-based learning to improve model performance. Defaults to 0, indicating no few-shot examples
        by default.
        - keys (Union[List[str], str], optional): A list of OpenAI API keys or a single OpenAI API key.
        - key_start_id (int, optional): The starting index for the OpenAI API key cycle.
        - proxy (str, optional): The proxy configuration for OpenAI API calls.
        - api_type (str, optional): The type of API service, if using Azure AI as the backend.
        - api_base (str, optional): The base URL for the API, applicable when using Azure AI.
        - api_version (str, optional): The API version, necessary for Azure AI integration.

        Raises:
        - ValueError: If an unsupported prompt mode is provided or if no OpenAI API keys / invalid OpenAI API keys are supplied.

        Note:
        - This class supports cycling between multiple OpenAI API keys to distribute quota usage or handle rate limiting.
        - Azure AI integration is depends on the presence of `api_type`, `api_base`, and `api_version`.
        """
        super().__init__(
            model, context_size, prompt_mode, max_output_tokens, num_few_shot_examples
        )
        if isinstance(keys, str):
            keys = [keys]
        if not keys:
            raise ValueError("Please provide OpenAI Keys.")
        if prompt_mode not in [PromptMode.CHATQA]:
            raise ValueError(
                f"unsupported prompt mode for GPT models: {prompt_mode}, expected {PromptMode.CHATQA}."
            )

        self._keys = keys
        self._cur_key_id = key_start_id or 0
        self._cur_key_id = self._cur_key_id % len(self._keys)
        self._post_processor = GPTPostProcessor()
        openai.proxy = proxy
        openai.api_key = self._keys[self._cur_key_id]
        self.use_azure_ai = False

        if all([api_type, api_base, api_version]):
            # See https://learn.microsoft.com/en-US/azure/ai-services/openai/reference for list of supported versions
            openai.api_version = api_version
            openai.api_type = api_type
            openai.api_base = api_base
            self.use_azure_ai = True

    class CompletionMode(Enum):
        UNSPECIFIED = 0
        CHAT = 1
        TEXT = 2

    def _call_completion(
        self,
        *args,
        completion_mode: CompletionMode,
        return_text=False,
        reduce_length=False,
        **kwargs,
    ) -> Union[str, Dict[str, Any]]:
        while True:
            try:
                if completion_mode == self.CompletionMode.CHAT:
                    completion = openai.chat.completions.create(
                        *args, **kwargs, timeout=30
                    )
                elif completion_mode == self.CompletionMode.TEXT:
                    completion = openai.Completion.create(*args, **kwargs)
                else:
                    raise ValueError(
                        "Unsupported completion mode: %V" % completion_mode
                    )
                break
            except Exception as e:
                print(str(e))
                if "This model's maximum context length is" in str(e):
                    print("reduce_length")
                    return "ERROR::reduce_length"
                if "The response was filtered" in str(e):
                    print("The response was filtered")
                    return "ERROR::The response was filtered"
                self._cur_key_id = (self._cur_key_id + 1) % len(self._keys)
                openai.api_key = self._keys[self._cur_key_id]
                time.sleep(0.1)
        if return_text:
            completion = (
                completion.choices[0].message.content
                if completion_mode == self.CompletionMode.CHAT
                else completion.choices[0].text
            )
        return completion

    def run_llm(
        self,
        prompt: Union[str, List[Dict[str, str]]],
        logging: bool = False,
    ) -> Tuple[str, RAGExecInfo]:
        model_key = "model"
        if logging:
            print(f"Prompt: {prompt}")
        response = self._call_completion(
            messages=prompt,
            temperature=0.1,
            completion_mode=SafeOpenai.CompletionMode.CHAT,
            return_text=True,
            **{model_key: self._model},
        )
        try:
            encoding = tiktoken.get_encoding(self._model)
        except:
            encoding = tiktoken.get_encoding("cl100k_base")
        if logging:
            print(f"Response: {response}")
        answers, rag_exec_response = self._post_processor(response)
        if logging:
            print(f"Answers: {answers}")
        rag_exec_info = RAGExecInfo(
            prompt=prompt,
            response=rag_exec_response,
            input_token_count=self.get_num_tokens(prompt),
            output_token_count=sum([len(ans.text) for ans in answers]),
            candidates=[],
        )
        if logging:
            print(f"RAG Exec Info: {rag_exec_info}")
        return answers, rag_exec_info

    def create_prompt(
        self, request: Request, topk: int
    ) -> Tuple[List[Dict[str, str]], int]:
        query = request.query.text
        max_length = (self._context_size - 200) // topk
        while True:
            rank = 0
            context = []
            for cand in request.candidates[:topk]:
                rank += 1
                content = self.convert_doc_to_prompt_content(cand.doc, max_length)
                context.append(
                    f"[{rank}] {self._replace_number(content)}",
                )
            if self._prompt_mode == PromptMode.CHATQA:
                chat_qa_template = ChatQATemplate()
                messages = chat_qa_template(query, context, "gpt")
            else:
                raise ValueError(
                    f"Unsupported prompt mode: {self._prompt_mode}, expected {PromptMode.CHATQA}."
                )
            num_tokens = self.get_num_tokens(messages)
            if num_tokens <= self.max_tokens() - self.num_output_tokens():
                break
            else:
                max_length -= max(
                    1,
                    (num_tokens - self.max_tokens() + self.num_output_tokens())
                    // (topk * 4),
                )
        return messages, self.get_num_tokens(messages)

    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """Returns the number of tokens used by a list of messages in prompt."""
        if self._model in ["gpt-3.5-turbo-0301", "gpt-3.5-turbo"]:
            tokens_per_message = (
                4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            )
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif self._model in ["gpt-4-0314", "gpt-4"]:
            tokens_per_message = 3
            tokens_per_name = 1
        else:
            tokens_per_message, tokens_per_name = 0, 0

        try:
            encoding = tiktoken.get_encoding(self._model)
        except:
            encoding = tiktoken.get_encoding("cl100k_base")

        num_tokens = 0
        if isinstance(prompt, list):
            for message in prompt:
                num_tokens += tokens_per_message
                for key, value in message.items():
                    num_tokens += len(encoding.encode(value))
                    if key == "name":
                        num_tokens += tokens_per_name
        else:
            num_tokens += len(encoding.encode(prompt))
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def cost_per_1k_token(self, input_token: bool) -> float:
        # Brought in from https://openai.com/pricing on 2023-07-30
        cost_dict = {
            ("gpt-3.5", 4096): 0.0015 if input_token else 0.002,
            ("gpt-3.5", 16384): 0.003 if input_token else 0.004,
            ("gpt-4", 8192): 0.03 if input_token else 0.06,
            ("gpt-4", 32768): 0.06 if input_token else 0.12,
        }
        model_key = "gpt-3.5" if "gpt-3" in self._model else "gpt-4"
        return cost_dict[(model_key, self._context_size)]



from typing import List

from ftfy import fix_text


class ChatQATemplate:
    def __init__(self):
        self.system_message_gpt = "This is a chat between a user and an artificial intelligence assistant. The assistant gives helpful and detailed answers to the user's question based on the context. The assistant should also indicate when the answer cannot be found in the context."
        self.system_message_osllm = "System: This is a chat between a user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions based on the context. The assistant should also indicate when the answer cannot be found in the context."
        self.input_context = "Context: {context}"
        self.instruction_official = (
            "Please give a full and complete answer for the question."
        )
        self.instruction_ragnarok = (
            "Please give a full and complete answer for the question. "
            + "Cite each context document inline that supports your answer within brackets [] using the IEEE format. "
            + "Ensure each sentence is properly cited."
        )
        self.user_input = "{query}"
        self.sep = "\n\n"

    def __call__(self, query: str, context: List[str], model: str) -> List[str]:
        str_context = self.sep.join(context)
        if "gpt" in model:
            user_input_context = (
                f"Instruction: {self.instruction_ragnarok}"
                + self.sep
                + f"Context: {self.input_context.format(context=str_context)}"
                + self.sep
                + f"Query: {query}"
                + self.sep
                + f"Instruction: {self.instruction_ragnarok}"
            )
            user_input_context += "\n\nAnswer:"
        else:
            user_input_context = (
                f"Instruction: {self.instruction_ragnarok}"
                + " "
                + f"Question: {query}{self.sep}Assistant:"
            )

        if "gpt" in model:
            messages = []
            messages.append(
                {
                    "role": "system",
                    "content": self.system_message_gpt,
                }
            )
            messages.append(
                {
                    "role": "user",
                    "content": fix_text(user_input_context),
                }
            )
            return messages
        else:
            prompt = f"{self.system_message_osllm}{self.sep}{self.input_context.format(context=str_context)}{self.sep}User: {user_input_context}"
        # else:
        #     conv = get_conversation_template(model)
        #     conv.set_system_message(self.system_message)
        #     conv.append_message(conv.roles[0], user_input_context)
        #     conv.append_message(conv.roles[1], None)
        #     prompt = conv.get_prompt()
        prompt = fix_text(prompt)
        return prompt



import argparse

from flask import Flask, jsonify

from ragnarok import retrieve_and_generate
from ragnarok.retrieve_and_rerank.retriever import RetrievalMethod


def create_app():
    app = Flask(__name__)

    @app.route(
        "/api/model/<string:model_path>/index/<string:dataset>/<string:host>/reranker/<string:reranker_method>/<string:host_reranker>/query=<string:query>&hits_retriever=<int:top_k_retrieve>&hits_reranker=<int:top_k_rerank>&qid=<int:qid>",
        methods=["GET"],
    )
    def search(
        model_path,
        dataset,
        host,
        reranker_method,
        host_reranker,
        query,
        top_k_retrieve,
        top_k_rerank,
        qid,
    ):
        try:
            if reranker_method == "rank_zephyr":
                retrieval_method = [RetrievalMethod.BM25, RetrievalMethod.RANK_ZEPHYR]
            elif reranker_method == "gpt_4o":
                retrieval_method = [RetrievalMethod.BM25, RetrievalMethod.RANK_GPT4O]
            else:
                return jsonify({"error": "Invalid reranker method"}), 400
            # Assuming the function is called with these parameters and returns a response
            response = retrieve_and_generate.retrieve_and_generate(
                dataset=dataset,
                retrieval_method=retrieval_method,
                query=query,
                model_path=model_path,
                host="http://localhost:" + host_reranker,
                interactive=True,
                k=[top_k_retrieve, top_k_rerank],
                qid=qid,
            )
            result = {
                "topic_id": response.query.qid,
                "topic": response.query.text,
                "references": response.references,
                "response_length": sum(
                    len(sentence["text"]) for sentence in response.answer
                ),
                "answer": [
                    {"text": sentence["text"], "citations": sentence["citations"]}
                    for sentence in response.answer
                ],
            }
            return jsonify(result), 200
        except Exception as e:
            return jsonify({"error": str(e)}), 500

    return app


def main():
    parser = argparse.ArgumentParser(description="Start the Ragnarok Flask server.")
    parser.add_argument(
        "--port", type=int, default=8084, help="The port to run the Flask server on."
    )

    args = parser.parse_args()

    app = create_app()
    app.run(host="0.0.0.0", port=args.port, debug=True)


if __name__ == "__main__":
    main()



from datetime import datetime
from pathlib import Path
from typing import List

from tqdm import tqdm

from ragnarok.data import DataWriter, OutputFormat, Request, Result
from ragnarok.generate.llm import LLM


class RAG:
    def __init__(self, agent: LLM) -> None:
        self._agent = agent

    def answer_batch(
        self,
        requests: List[Request],
        topk: int = 20,
        shuffle_candidates: bool = False,
        logging: bool = False,
    ) -> List[Result]:
        """
        Generates a list of attributed answers using the Ragnarok agent.

        Args:
            requests (List[Request]): The list of requests. Each request has a query and a candidates list.
            topk (int, optional): The end rank for processing. Defaults to 20.
            shuffle_candidates (bool, optional): Whether to shuffle candidates before answering. Defaults to False.
            logging (bool, optional): Enables logging of the answering process. Defaults to False.

        Returns:
            List[Result]: A list containing the attributed answers.
        """
        results = []
        for request in tqdm(requests):
            result = self._agent.answer_batch(
                [request],
                topk=min(topk, len(request.candidates)),
                shuffle_candidates=shuffle_candidates,
                logging=logging,
            )
            results.append(result[0])
        return results

    def answer(
        self,
        request: Request,
        rank_start: int = 0,
        rank_end: int = 20,
        shuffle_candidates: bool = False,
        logging: bool = False,
    ) -> Result:
        """
        Generates an attributed answer using the Ragnarok agent.

        Args:
            request (Request): The answering request which has a query and a candidates list.
            rank_start (int, optional): The starting rank for processing. Defaults to 0.
            rank_end (int, optional): The end rank for processing. Defaults to 20.
            shuffle_candidates (bool, optional): Whether to shuffle candidates before answering. Defaults to False.
            logging (bool, optional): Enables logging of the answering process. Defaults to False.

        Returns:
            Result: the generated result which contains the attributed answer.
        """
        results = self.answer_batch(
            requests=[request],
            rank_start=rank_start,
            rank_end=rank_end,
            shuffle_candidates=shuffle_candidates,
            logging=logging,
        )
        return results[0]

    def write_answer_results(
        self,
        retrieval_method_name: str,
        results: List[Result],
        shuffle_candidates: bool = False,
        top_k_candidates: int = 20,
        dataset_name: str = None,
        results_dirname: str = "results",
        rag_execution_summary_dirname: str = "rag_execution_summary",
        output_format: OutputFormat = OutputFormat.JSONL,
    ) -> str:
        """
        Writes the attributed answers to files in specified formats.

        This function saves the results only in the JSON format expected in TREC 2024 RAG.

        Args:
            retrieval_method_name (str): The name of the retrieval method.
            results (List[Result]): The results to be written.
            shuffle_candidates (bool, optional): Indicates if the candidates were shuffled. Defaults to False.
            top_k_candidates (int, optional): The number of top candidates considered. Defaults to 20.
            dataset_name (str, optional): The name of the dataset used. Defaults to None.
            results_dirname (str, optional): The directory name to save the results. Defaults to "results".
            rag_execution_summary_dirname (str, optional): The directory name to save the RAG execution summary. Defaults to "rag_execution_summary".
            output_format (OutputFormat, optional): The output format to save the results. Defaults to OutputFormat.JSONL.

        Returns:
            str: The file name of the saved results.

        Note:
            The function creates directories and files as needed. The file names are constructed based on the
            provided parameters and the current timestamp to ensure uniqueness so there are no collisions.
        """
        _modelname = self._agent._model.split("/")[-1]
        if _modelname.startswith("checkpoint"):
            _modelname = self._agent._model.split("/")[-2] + "_" + _modelname
        name = f"{_modelname}_{self._agent._context_size}_{top_k_candidates}_{self._agent._prompt_mode}"
        if dataset_name:
            name = f"{name}_{dataset_name}"
        if self._agent._num_few_shot_examples > 0:
            name += f"_{self._agent._num_few_shot_examples}_shot"
        name = (
            f"{name}_shuffled_{datetime.isoformat(datetime.now())}"
            if shuffle_candidates
            else f"{name}_{datetime.isoformat(datetime.now())}"
        )
        # write generate results
        writer = DataWriter(results)
        Path(f"{results_dirname}/{retrieval_method_name}/").mkdir(
            parents=True, exist_ok=True
        )
        if output_format == OutputFormat.JSON:
            output_file = f"{results_dirname}/{retrieval_method_name}/{name}.json"
            writer.write_in_json_format(output_file)
        else:
            output_file = f"{results_dirname}/{retrieval_method_name}/{name}.jsonl"
            writer.write_in_jsonl_format(output_file)
        Path(f"{rag_execution_summary_dirname}/{retrieval_method_name}/").mkdir(
            parents=True, exist_ok=True
        )
        writer.write_rag_exec_summary(
            f"{rag_execution_summary_dirname}/{retrieval_method_name}/{name}.jsonl"
        )
        return output_file



import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Union

from dacite import from_dict


@dataclass
class Query:
    text: str
    qid: Union[str | int]


@dataclass
class Candidate:
    docid: Union[str | int]
    score: float
    doc: Dict[str, Any]


@dataclass
class Request:
    query: Query
    candidates: List[Candidate] = field(default_factory=list)
    # Optional Ranking Exec Summary
    ranking_exec_summary: List[Dict[str, Any]] = field(default_factory=list)


@dataclass
class RAGExecInfo:
    prompt: Any
    response: Any
    input_token_count: int
    output_token_count: int
    candidates: List[Candidate] = field(default_factory=list)


@dataclass
class CitedSentence:
    text: str
    citations: List[int] = field(default_factory=list)


@dataclass
class Result:
    query: Query
    references: List[Union[str | int]] = field(default_factory=list)
    answer: List[CitedSentence] = field(default_factory=list)
    rag_exec_summary: RAGExecInfo = None


class OutputFormat(Enum):
    JSON = "json"
    JSONL = "jsonl"

    def __str__(self):
        return self.value


def read_requests_from_file(file_path: str) -> List[Request]:
    extension = file_path.split(".")[-1]
    if extension == "jsonl":
        requests = []
        with open(file_path, "r") as f:
            for l in f:
                if not l.strip():
                    continue
                requests.append(from_dict(data_class=Request, data=json.loads(l)))
        return requests
    elif extension == "json":
        with open(file_path, "r") as f:
            request_dicts = json.load(f)
        return [
            from_dict(data_class=Request, data=request_dict)
            for request_dict in request_dicts
        ]
    else:
        raise ValueError(f"Expected json or jsonl file format, got {extension}")


def read_results_from_file(file_path: str) -> List[Result]:
    extension = file_path.split(".")[-1]
    if extension == "jsonl":
        results = []
        with open(file_path, "r") as f:
            for l in f:
                if not l.strip():
                    continue
                result_dict = json.loads(l)
                result = Result(
                    query=Query(text=result_dict["topic"], qid=result_dict["topic_id"]),
                    references=result_dict["references"],
                    answer=[
                        CitedSentence(
                            text=sentence["text"], citations=sentence["citations"]
                        )
                        for sentence in result_dict["answer"]
                    ],
                    rag_exec_summary=None,
                )
                results.append(result)
        return results
    elif extension == "json":
        with open(file_path, "r") as f:
            result_dicts = json.load(f)
        return [
            Result(
                query=Query(text=result_dict["topic"], qid=result_dict["topic_id"]),
                references=result_dict["references"],
                answer=[
                    CitedSentence(
                        text=sentence["text"], citations=sentence["citations"]
                    )
                    for sentence in result_dict["answer"]
                ],
                rag_exec_summary=None,
            )
            for result_dict in result_dicts
        ]
    else:
        raise ValueError(f"Expected json or jsonl file format, got {extension}")


def remove_unused_references(result: Result, max_per_sentence: int = 3) -> Result:
    # Find all referenced document ids in the citations
    cited_docids = set()
    for cited_sentence in result.answer:
        cited_sentence.citations = cited_sentence.citations[:max_per_sentence]
        cited_docids.update(cited_sentence.citations)

    # Filter the references list to only include cited docids
    result.references = [
        ref for i, ref in enumerate(result.references) if i in cited_docids
    ]

    # Create a mapping from old indices to new indices
    new_index_map = {
        old_idx: new_idx for new_idx, old_idx in enumerate(sorted(cited_docids))
    }

    # Update citations in CitedSentence
    for cited_sentence in result.answer:
        cited_sentence.citations = [
            new_index_map[old_idx] for old_idx in cited_sentence.citations
        ]

    return result


class DataWriter:
    def __init__(
        self,
        data: Union[Request | Result | List[Result] | List[Request]],
        append: bool = False,
    ):
        if isinstance(data, list):
            self._data = data
        else:
            self._data = [data]
        self._append = append

    def write_rag_exec_summary(self, filename: str):
        exec_summary = []
        with open(filename, "a" if self._append else "w") as f:
            for d in self._data:
                exec_summary = {
                    "query": d.query.__dict__,
                    "rag_exec_summary": d.rag_exec_summary.__dict__,
                }
                f.write(json.dumps(exec_summary) + "\n")

    def _convert_result_to_dict(self, result: Result) -> Dict:
        result_dict = {
            "topic_id": result.query.qid,
            "topic": result.query.text,
            "references": result.references,
            "response_length": sum(len(sentence.text) for sentence in result.answer),
            "answer": [
                {"text": sentence.text, "citations": sentence.citations}
                for sentence in result.answer
            ],
        }
        return result_dict

    def write_in_json_format(self, filename: str):
        formatted_data = [self._convert_result_to_dict(d) for d in self._data]
        with open(filename, "a" if self._append else "w") as f:
            json.dump(formatted_data, f, indent=2)

    def write_in_jsonl_format(self, filename: str):
        with open(filename, "a" if self._append else "w") as f:
            for d in self._data:
                json.dump(self._convert_result_to_dict(d), f)
                f.write("\n")



HITS_INFO = {
    # TODO(ronak): Add cached results from retrieval methods
    # BM25
    # "BM25/retrieve_results_researchy_questions_test_top100.json": {
    #     "description": "Top-100 BM25 Results for a subset of the Researchy Questions task. (Cached JSON format)",
    #     "urls": [
    #         "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_covid_top100.json"
    #     ],
    #     "md5": "84ee088e57c6bd8e274b25ce45406317",
    #     "size (bytes)": 9113367,
    #     "downloaded": False,
    # },
    # RANK_ZEPHYR
    # "RANK_ZEPHYR/retrieve_results_covid_top100.json": {
}



import os
import sys
from pathlib import Path

from dacite import from_dict

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
parent = os.path.dirname(SCRIPT_DIR)
parent = os.path.dirname(parent)
sys.path.append(parent)

from ragnarok.data import DataWriter, Request
from ragnarok.generate.cohere import Cohere
from ragnarok.generate.gpt import GPT

request_dict = {
    "query": {"text": "how long is life cycle of flea", "qid": "264014"},
    "candidates": [
        {
            "doc": {
                "segment": "The life cycle of a flea can last anywhere from 20 days to an entire year. It depends on how long the flea remains in the dormant stage (eggs, larvae, pupa). Outside influences, such as weather, affect the flea cycle. A female flea can lay around 20 to 25 eggs in one day."
            },
            "docid": "4834547",
            "score": 14.971799850463867,
        },
        {
            "doc": {
                "segment": "The life cycle of a flea can last anywhere from 20 days to an entire year. It depends on how long the flea remains in the dormant stage (eggs, larvae, pupa). Outside influences, such as weather, affect the flea cycle. A female flea can lay around 20 to 25 eggs in one day. The flea egg stage is the beginning of the flea cycle. This part of the flea cycle represents a little more than one third of the flea population at any given time. Depending on the temperature and humidity of the environment the egg can take from two to six days to hatch."
            },
            "docid": "6641238",
            "score": 15.090800285339355,
        },
        {
            "doc": {
                "segment": "To go to our more detailed flea life cycle and flea control page, click here. 1) The flea life cycle diagram - a complete step-by-step diagram of animal host flea infestation; flea reproduction and environmental flea contamination with juvenile flea life cycle stages (eggs, larvae and pupae)."
            },
            "docid": "1610712",
            "score": 13.455499649047852,
        },
        {
            "doc": {
                "segment": "Flea Pupa. The flea larvae spin cocoons around themselves in which they move to the last phase of the flea life cycle and become adult fleas. The larvae can remain in the cocoon anywhere from one week to one year. Temperature is one factor that determines how long it will take for the adult flea to emerge from the cocoon."
            },
            "docid": "96852",
            "score": 14.215100288391113,
        },
        {
            "doc": {
                "segment": "The cat flea's primary host is the domestic cat, but it is also the primary flea infesting dogs in most of the world. The cat flea can also maintain its life cycle on other carnivores and on omnivores. Humans can be bitten, though a long-term population of cat fleas cannot be sustained and infest people.However, if the female flea is allowed to feed for 12 consecutive hours on a human, it can lay viable eggs.he cat flea's primary host is the domestic cat, but it is also the primary flea infesting dogs in most of the world. The cat flea can also maintain its life cycle on other carnivores and on omnivores. Humans can be bitten, though a long-term population of cat fleas cannot be sustained and infest people."
            },
            "docid": "4239616",
            "score": 13.947500228881836,
        },
        {
            "doc": {
                "segment": "5. Cancel. A flea can live up to a year, but its general lifespan depends on its living conditions, such as the availability of hosts. Find out how long a flea's life cycle can last with tips from a pet industry specialist in this free video on fleas and pest control.Part of the Video Series: Flea Control.ancel. A flea can live up to a year, but its general lifespan depends on its living conditions, such as the availability of hosts. Find out how long a flea's life cycle can last with tips from a pet industry specialist in this free video on fleas and pest control. Part of the Video Series: Flea Control."
            },
            "docid": "5611210",
            "score": 15.780599594116211,
        },
        {
            "doc": {
                "segment": "2) The fleas life cycle discussed - the flea life cycle diagram explained in full. 2a) Fleas life cycle 1 - The adult flea lays her eggs on the host animal. 2b) Fleas life cycle 2 - The egg falls off the animal's skin and into the local environment of the host animal. 2c) Fleas life cycle 3 - The flea egg hatches, releasing a first stage (stage 1) flea larva."
            },
            "docid": "96854",
            "score": 13.985199928283691,
        },
        {
            "doc": {
                "segment": "In appearance, flea larvae can be up to \u00c2\u00bc-inch long and are white (almost see-through) and legless. Larvae make up about 35 percent of the flea population in the average household. If conditions are favorable, the larvae will spin cocoons in about 5-20 days of hatching from their eggs.This leads to the next life stage, called the cocoon or pupae stage.The pupae stage of the flea life cycle accounts for about 10 percent of the flea population in a home.f conditions are favorable, the larvae will spin cocoons in about 5-20 days of hatching from their eggs. This leads to the next life stage, called the cocoon or pupae stage. The pupae stage of the flea life cycle accounts for about 10 percent of the flea population in a home."
            },
            "docid": "5635521",
            "score": 13.533599853515625,
        },
    ],
}

request = from_dict(data_class=Request, data=request_dict)
rag_cohere = Cohere()
rag_cohere_results = rag_cohere.answer(request=request)
rag_gpt4 = GPT("gpt-4")
rag_gpt4_results = rag_gpt4.answer(request=request)

print("Cohere Results:")
for result in rag_cohere_results:
    print(result)

print("\nGPT-4 Results:")
for result in rag_gpt4_results:
    print(result)

# write rerank results
writer = DataWriter(rag_cohere_results)
Path(f"demo_outputs/").mkdir(parents=True, exist_ok=True)
writer.write_in_json_format(f"demo_outputs/rag_results.json")



import os
import re
from typing import Any, Dict, List, Tuple

import spacy
import stanza

from ragnarok.data import CitedSentence


class StanzaTokenizer:
    def __init__(self, lang="en", processors="tokenize"):
        self.pipeline = stanza.Pipeline(lang=lang, processors=processors)

    def tokenize(
        self, text: str, sep: str = "\n", strip: List[str] = ["-"]
    ) -> List[str]:
        """
        Tokenize the input text into sentences.
        Args:
            text: str: input text
            sep: str: separator to split sentences
            strip: List[str]: list of strings to strip from the text

        Returns:
            List[str]: list of sentences
        """
        new_sentences = []
        for strip_str in strip:
            text = text.replace(strip_str, "")
        for sentence in self.pipeline(text).sentences:
            if sep in sentence.text:
                new_sentences += [text.strip() for text in sentence.text.split(sep)]
            else:
                new_sentences.append(sentence.text.strip())
        return new_sentences


class SpacyTokenizer:
    def __init__(self, model="en_core_web_trf"):
        try:
            self.nlp = spacy.load(model)
        except:
            os.system(f"python -m spacy download {model}")
            self.nlp = spacy.load(model)

    def tokenize(self, text: str, replace_newline: str = " ") -> List[str]:
        """
        Tokenize the input text into sentences.
        Args:
            text: str: input text
            replace_newline: str: replace newline character with this character

        Returns:
            List[str]: list of sentences
        """
        sentences = []
        text = replace_newline.join(
            text.replace("\n", replace_newline).split(replace_newline)
        )
        for sent in self.nlp(text).sents:
            if re.search("[a-zA-Z]", sent.text):
                sentences.append(sent.text)
        return sentences


class CoherePostProcessor:
    def __init__(self, tokenizer="spacy") -> None:
        self.tokenizer = (
            StanzaTokenizer(lang="en", processors="tokenize")
            if tokenizer == "stanza"
            else SpacyTokenizer(model="en_core_web_trf")
        )

    def _find_sentence_citations(
        self, text_output: str, sentence: str, cohere_citations: List[Dict[str, Any]]
    ) -> List[int]:
        start_pos = text_output.find(sentence)
        end_pos = start_pos + len(sentence)
        citations = []
        for citation in cohere_citations:
            if citation["start"] >= start_pos and citation["end"] <= end_pos:
                citations += citation["document_ids"]
            elif (
                citation["start"] < start_pos
                and citation["end"] <= end_pos
                and citation["end"] > start_pos
            ):
                citations += citation["document_ids"]
            elif (
                citation["start"] >= start_pos
                and citation["start"] < end_pos
                and citation["end"] > end_pos
            ):
                citations += citation["document_ids"]
            elif citation["start"] < start_pos and citation["end"] > end_pos:
                citations += citation["document_ids"]

        if citations:
            citations = [
                int(doc_id.replace("doc_", "")) for doc_id in list(set(citations))
            ]
            citations = sorted(citations)
        return citations

    def __call__(self, response) -> Tuple[List[CitedSentence], Dict[str, Any]]:
        text_output = response.text
        if not response.citations:
            citations = []
        else:
            citations = [
                {
                    "start": citation.start,
                    "end": citation.end,
                    "text": citation.text,
                    "document_ids": list(citation.document_ids),
                }
                for citation in response.citations
            ]
        rag_exec_response = {"text": response.text, "citations": citations}
        sentences = self.tokenizer.tokenize(text_output)
        answers = []
        for sentence in sentences:
            answer_citations = self._find_sentence_citations(
                text_output, sentence, citations
            )
            answers.append(CitedSentence(text=sentence, citations=answer_citations))

        return answers, rag_exec_response


class GPTPostProcessor:
    def __init__(self, tokenizer="spacy") -> None:
        self.tokenizer = SpacyTokenizer(model="en_core_web_trf")

    def _find_sentence_citations(
        self, sentence: str, citation_range: List[int] = list(range(20))
    ) -> tuple[str, List[int]]:
        # Regex pattern to find citations
        pattern = re.compile(r"\[\d+\](?:,? ?)")

        # Find all citations
        citations = pattern.findall(sentence)
        if citations:
            # Remove citations from text
            sentence = pattern.sub("", sentence).strip()

            # Extract indices from citations
            indices = [
                int(re.search(r"\d+", citation).group()) - 1 for citation in citations
            ]
            citations = [index for index in indices if index in citation_range]
        else:
            matches = re.findall(r"\[[^\]]*\]", sentence)
            if not matches:
                return sentence, []
            citations = []
            for match in matches:
                citation = match[1:-1]
                try:
                    if "," in citation:
                        flag = False
                        for cit in citation.split(","):
                            cit = int(cit) - 1
                            if cit in citation_range:
                                flag = True
                                citations.append(int(cit))
                        if flag:
                            sentence = sentence.replace(match, "")
                    else:
                        citation = int(citation) - 1
                        if citation in citation_range:
                            citations.append(citation)
                            sentence = sentence.replace(match, "")
                except:
                    print(f"Not a valid citation: {match}")

        sentence = re.sub(" +", " ", sentence)
        if len(sentence) > 3:
            if sentence[-2] == " ":
                sentence = sentence[:-2] + sentence[-1]
        return sentence, citations

    def __call__(self, response) -> List[Dict[str, Any]]:
        text_output = response
        sentences = self.tokenizer.tokenize(text_output)
        answers = []
        citation_range = list(range(20))
        for sentence in sentences:
            sentence_parsed, citations = self._find_sentence_citations(
                sentence, citation_range
            )
            answers.append(CitedSentence(text=sentence_parsed, citations=citations))

        rag_exec_response = {"text": response, "citations": citation_range}

        return answers, rag_exec_response



import json
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Union

from dacite import from_dict

from ragnarok.data import Request


class RetrievalMode(Enum):
    DATASET = "dataset"
    CUSTOM = "custom"

    def __str__(self):
        return self.value


class CacheInputFormat(Enum):
    JSON = "json"
    JSONL = "jsonl"

    def __str__(self):
        return self.value


class RetrievalMethod(Enum):
    UNSPECIFIED = "unspecified"
    BM25 = "bm25"
    RANK_ZEPHYR = "rank_zephyr"
    RANK_ZEPHYR_RHO = "rank_zephyr_rho"
    RANK_VICUNA = "rank_vicuna"
    RANK_GPT4O = "gpt-4o"
    RANK_GPT4 = "gpt-4"
    RANK_GPT35_TURBO = "gpt-3.5-turbo"

    def __str__(self):
        return self.value

    @staticmethod
    def from_string(value):
        for method in RetrievalMethod:
            if method.value == value:
                return method
        raise ValueError(f"Unknown retrieval method: {value}")


class Retriever:
    def __init__(
        self,
        retrieval_mode: RetrievalMode,
        dataset: Union[str, List[str], List[Dict[str, Any]]],
        retrieval_method: Union[
            RetrievalMethod, List[RetrievalMethod]
        ] = RetrievalMethod.UNSPECIFIED,
        query: str = None,
        index_path: str = None,
        topics_path: str = None,
    ) -> None:
        self._retrieval_mode = retrieval_mode
        self._dataset = dataset
        self._retrieval_method = (
            retrieval_method
            if isinstance(retrieval_method, list)
            else [retrieval_method]
        )
        self._query = query
        self._index_path = index_path
        self._topics_path = topics_path

    @staticmethod
    def from_dataset_with_prebuilt_index(
        dataset_name: str,
        retrieval_method: Union[
            RetrievalMethod, List[RetrievalMethod]
        ] = RetrievalMethod.BM25,
        cache_input_format: CacheInputFormat = CacheInputFormat.JSONL,
        k: List[int] = [100, 20],
    ):
        """
        Creates a Retriever instance for a dataset with a prebuilt index.

        Args:
            dataset_name (str): The name of the dataset.
            retrieval_method (Union[RetrievalMethod, List[RetrievalMethod]]): The retrieval method(s) to be used. Defaults to BM25.
            k (List[int], optional): The top k hits to retrieve. Defaults to [100, 20].

        Returns:
            List[Request]: The list of requests. Each request has a query and list of candidates.

        Raises:
            ValueError: If dataset name or retrieval method is invalid or missing.
        """
        if not dataset_name:
            raise ValueError("Please provide name of the dataset.")
        if not isinstance(dataset_name, str):
            raise ValueError(
                f"Invalid dataset format: {dataset_name}. Expected a string representing name of the dataset."
            )
        if not retrieval_method:
            raise "Please provide a retrieval method."
        if RetrievalMethod.UNSPECIFIED in retrieval_method:
            raise ValueError(
                f"Invalid retrieval method: {retrieval_method}. Please provide a specific retrieval method."
            )
        retriever = Retriever(
            RetrievalMode.DATASET,
            dataset=dataset_name,
            retrieval_method=retrieval_method,
        )
        return retriever.retrieve(k=k, cache_input_format=cache_input_format)

    def retrieve(
        self,
        retrieve_results_dirname: str = "retrieve_results",
        cache_input_format: CacheInputFormat = CacheInputFormat.JSONL,
        k: List[int] = [100, 20],
    ) -> List[Request]:
        """
        Executes the retrieval process based on the configuration provided with the Retriever instance.

        Returns:
            List[Request]: The list of requests. Each request has a query and list of candidates.

        Raises:
            ValueError: If the retrieval mode is invalid or the format is not as expected.
        """
        if self._retrieval_mode == RetrievalMode.DATASET:
            candidates_file = Path(
                f"{retrieve_results_dirname}/{self._retrieval_method[-1].name}/retrieve_results_{self._dataset}_top{k[-1]}.{cache_input_format}"
            )
            print(f"Loading candidates from {candidates_file}.")
            if not candidates_file.is_file():
                # Get top100 file instead
                try:
                    candidates_file = Path(
                        f"{retrieve_results_dirname}/{self._retrieval_method[-1].name}/retrieve_results_{self._dataset}_top100.{cache_input_format}"
                    )
                    print()
                    if not candidates_file.is_file():
                        raise ValueError(f"File not found: {candidates_file}")
                except ValueError as e:
                    print(f"Failed to load JSON file: {candidates_file}")
            if candidates_file.is_file():
                if cache_input_format == CacheInputFormat.JSON:
                    with open(candidates_file, "r") as f:
                        loaded_results = json.load(f)
                elif cache_input_format == CacheInputFormat.JSONL:
                    with open(candidates_file, "r") as f:
                        loaded_results = [json.loads(l) for l in f]
                retrieved_results = [
                    from_dict(data_class=Request, data=r) for r in loaded_results
                ]
                # TODO remove!!! Filter those if query.qid has _ and does not end with _0
                retrieved_results = [
                    r
                    for r in retrieved_results
                    # if "_" in r.query.qid and not r.query.qid.endswith("_0")
                ]
                # Ensure the candidates are of at most k length
                for r in retrieved_results:
                    r.candidates = r.candidates[: k[-1]]
                print(
                    f"Loaded {len(retrieved_results)} requests from {candidates_file}."
                )
        else:
            raise ValueError(f"Invalid retrieval mode: {self._retrieval_mode}")
        return retrieved_results



import concurrent.futures
import os
import random

import gradio as gr

from ragnarok import retrieve_and_generate
from ragnarok.retrieve_and_rerank.retriever import RetrievalMethod

retriever_options = ["bm25"]
reranker_options = ["rank_zephyr", "rank_vicuna", "gpt-4o", "unspecified"]
llm_options = ["command-r", "command-r-plus", "gpt-4o", "gpt-35-turbo", "gpt-4"]


def generate_text_with_citations(response):
    output = []
    citation_texts = response.rag_exec_summary.candidates
    for sentence in response.answer:
        text = sentence.text
        citations = sentence.citations
        if citations:
            citation_html = ""
            for citation in citations:
                citation_title = citation_texts[citation]["doc"]["title"]
                citation_text = citation_texts[citation]["doc"]["segment"]
                citation_url = citation_texts[citation]["doc"]["url"]
                citation_html += f' \
<span class="tooltip">[{citation}] \
    <span class="tooltip-body"> \
        <h3>{citation_title}</h3> \
        <br/> \
        <p>{citation_text}</p> \
        <br/> \
        <a href="{citation_url}">{citation_url}</a> \
    </span> \
</span> '
            text += f" {citation_html}"
        output.append(text)
    return "<br/><br/>".join(output)


def query_model(
    retriever_path,
    reranker_path,
    LLM,
    dataset,
    host_retriever,
    host_reranker,
    top_k_retrieve,
    top_k_rerank,
    qid,
    query,
    num_passes=1,
):
    # RetreivalMethod Options:
    # UNSPECIFIED = "unspecified"
    # BM25 = "bm25"
    # RANK_ZEPHYR = "rank_zephyr"
    # RANK_ZEPHYR_RHO = "rank_zephyr_rho"
    # RANK_VICUNA = "rank_vicuna"
    # RANK_GPT4O = "gpt-4o"
    # RANK_GPT4 = "gpt-4"
    # RANK_GPT35_TURBO = "gpt-3.5-turbo"

    try:
        retriever_path = RetrievalMethod.from_string(retriever_path.lower())
        reranker_path = RetrievalMethod.from_string(reranker_path.lower())
    except KeyError:
        retriever_path = RetrievalMethod.UNSPECIFIED
        reranker_path = RetrievalMethod.UNSPECIFIED

    retrieval_method = [retriever_path, reranker_path]

    # Rerank method can be none (RetrievalMethod.UNSPECIFIED)
    if retrieval_method[0] == RetrievalMethod.UNSPECIFIED:
        raise ValueError(
            f"Invalid retrieval method: {retrieval_method}. Please provide a specific retrieval method."
        )

    try:
        response = retrieve_and_generate.retrieve_and_generate(
            dataset=dataset,
            query=query,
            host_reranker=host_reranker,
            host_retriever=host_retriever,
            interactive=True,
            k=[top_k_retrieve, top_k_rerank],
            qid=qid,
            retrieval_method=retrieval_method,
            generator_path=LLM,
            use_azure_openai=True,
            num_passes=num_passes,
        )
        output = generate_text_with_citations(response)
        result = {
            "topic_id": response.query.qid,
            "topic": response.query.text,
            "references": response.references,
            "response_length": sum(len(sentence.text) for sentence in response.answer),
            "answer": [
                {"text": sentence.text, "citations": sentence.citations}
                for sentence in response.answer
            ],
        }
        return [output, result]
    except Exception as e:
        return ["ERROR: " + str(e), None]


def on_submit(
    model_a,
    model_b,
    retriever_a,
    retriever_b,
    reranker_a,
    reranker_b,
    dataset,
    host_retriever_a,
    host_reranker_a,
    host_retriever_b,
    host_reranker_b,
    top_k_retrieve,
    top_k_rerank,
    qid,
    query,
    num_passes_a,
    num_passes_b,
    randomize,
):
    def query_wrapper(
        retriever, reranker, model, host_retriever, host_reranker, num_passes
    ):
        return query_model(
            retriever,
            reranker,
            model,
            dataset,
            host_retriever,
            host_reranker,
            top_k_retrieve,
            top_k_rerank,
            qid,
            query,
            num_passes,
        )

    executor = concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count())
    with executor:
        if randomize:
            retriever_a = random.choice(retriever_options)
            reranker_a = random.choice(reranker_options)
            model_a = random.choice(llm_options)
            retriever_b = random.choice(retriever_options)
            reranker_b = random.choice(reranker_options)
            model_b = random.choice(llm_options)

        futureA = executor.submit(
            query_wrapper,
            retriever_a,
            reranker_a,
            model_a,
            host_retriever_a,
            host_reranker_a,
            num_passes_a,
        )
        futureB = executor.submit(
            query_wrapper,
            retriever_b,
            reranker_b,
            model_b,
            host_retriever_b,
            host_reranker_b,
            num_passes_b,
        )

        resultA, responseA = futureA.result()
        resultB, responseB = futureB.result()

    return [resultA, resultB, responseA, responseB]


def on_submit_side_by_side(
    model_a,
    model_b,
    retriever_a,
    retriever_b,
    reranker_a,
    reranker_b,
    dataset,
    host_retriever_a,
    host_reranker_a,
    host_retriever_b,
    host_reranker_b,
    top_k_retrieve,
    top_k_rerank,
    qid,
    query,
    num_passes_a,
    num_passes_b,
):
    return on_submit(
        model_a,
        model_b,
        retriever_a,
        retriever_b,
        reranker_a,
        reranker_b,
        dataset,
        host_retriever_a,
        host_reranker_a,
        host_retriever_b,
        host_reranker_b,
        top_k_retrieve,
        top_k_rerank,
        qid,
        query,
        num_passes_a,
        num_passes_b,
        False,
    )


def on_submit_side_by_side_blinded(
    dataset,
    host_retriever_a,
    host_reranker_a,
    host_retriever_b,
    host_reranker_b,
    top_k_retrieve,
    top_k_rerank,
    qid,
    query,
    num_passes_a,
    num_passes_b,
):
    return on_submit(
        "",
        "",
        "",
        "",
        "",
        "",
        dataset,
        host_retriever_a,
        host_reranker_a,
        host_retriever_b,
        host_reranker_b,
        top_k_retrieve,
        top_k_rerank,
        qid,
        query,
        num_passes_a,
        num_passes_b,
        True,
    )


tooltip_style = """
<style>
.tooltip {
    position: relative;
    display: inline-block;
    cursor: pointer;
    text-decoration: underline;
}

.tooltip-body {
    visibility: hidden;
    position: absolute;
    z-index: 50;
    top: 100%;
    transform: translateX(-50%);
    width: 600px;
    color: #fff;
    background-color: rgba(0, 0, 0, 0.9);
    padding: 5px;
    border-radius: 5px;
    border: 1px solid rgba(255, 255, 255, 0.2);
    white-space: normal;
    word-wrap: break-word;
    overflow-wrap: break-word;
    box-shadow: 0px 4px 6px rgba(0, 0, 0, 0.1);
}

.tooltip:hover .tooltip-body {
  visibility: visible;
}

</style>
"""

html_content = """
<div class='navbar'>
    <h1>Ragnarök Chatbot Arena</h1>
    <p>Ask any question to RAG pipelines! Heavily built on the code for <a href="https://chat.lmsys.org">https://chat.lmsys.org</a> :)</p>
</div>
"""

retriever_options = ["bm25"]
reranker_options = ["rank_zephyr", "rank_vicuna", "gpt-4o", "unspecified"]
llm_options = ["command-r", "command-r-plus", "gpt-4o", "gpt-35-turbo", "gpt-4"]


def rag_pipeline_block(
    label_suffix="",
    default_retriever="bm25",
    default_reranker="rank_zephyr",
    default_llm="command-r",
):
    with gr.Column():
        retriever = gr.Dropdown(
            label=f"Retriever {label_suffix}",
            choices=retriever_options,
            value=default_retriever,
        )
        reranker = gr.Dropdown(
            label=f"Reranker {label_suffix}",
            choices=reranker_options,
            value=default_reranker,
        )
        llm = gr.Dropdown(
            label=f"LLM {label_suffix}", choices=llm_options, value=default_llm
        )

    return [retriever, reranker, llm]


def input_block():
    with gr.Row():
        input_text = gr.Textbox(
            label="Enter your query and press ENTER",
            placeholder="Type here...",
            value="What caused the second world war?",
        )
    with gr.Row():
        button = gr.Button("Compare")
    return [input_text, button]


def input_block_direct():
    with gr.Row():
        input_text = gr.Textbox(
            label="Enter your query and press ENTER",
            placeholder="Type here...",
            value="What caused the second world war?",
        )
    with gr.Row():
        button = gr.Button("Answer")
    return [input_text, button]


def output_block(side_by_side=True):
    with gr.Tab("Output"):
        with gr.Row():
            if side_by_side:
                pretty_output_a = gr.HTML(label="Pretty Output from Model A")
                pretty_output_b = gr.HTML(label="Pretty Output from Model B")
            else:
                pretty_output = gr.HTML(label="Pretty Output")
    with gr.Tab("Responses"):
        with gr.Row():
            if side_by_side:
                json_output_a = gr.JSON(label="JSON Output from Model A")
                json_output_b = gr.JSON(label="JSON Output from Model B")
            else:
                json_output = gr.JSON(label="JSON Output")

    if side_by_side:
        return [pretty_output_a, pretty_output_b, json_output_a, json_output_b]
    else:
        return [pretty_output, json_output]


def comparison_block():
    with gr.Row():
        a_better_button = gr.Button("👈 A is better")
        b_better_button = gr.Button("👉 B is better")
        both_good_button = gr.Button("🤝 Tie")
        both_bad_button = gr.Button("👎 Both are bad")

    return [a_better_button, b_better_button, both_good_button, both_bad_button]


def parameters_block(side_by_side=True):
    with gr.Accordion(label="Parameters", open=False):
        with gr.Column():
            dataset = gr.Dropdown(
                label="Dataset",
                choices=["msmarco-v2.1-doc-segmented"],
                value="msmarco-v2.1-doc-segmented",
            )
            top_k_retrieve = gr.Number(label="Hits Retriever", value=40)
            top_k_rerank = gr.Number(label="Hits Reranker", value=20)
            if side_by_side:
                with gr.Row():
                    host_retriever_a = gr.Textbox(
                        label="Retriever Host A", value="8081"
                    )
                    host_retriever_b = gr.Textbox(
                        label="Retriever Host B", value="8081"
                    )
                with gr.Row():
                    host_reranker_a = gr.Textbox(label="Reranker Host A", value="8082")
                    host_reranker_b = gr.Textbox(label="Reranker Host B", value="8083")
            else:
                host_retriever = gr.Textbox(label="Retriever Host", value="8081")
                host_reranker = gr.Textbox(label="Reranker Host", value="8082")
            if side_by_side:
                with gr.Row():
                    num_passes_a = gr.Textbox(
                        label="Number of rerank passes A", value="1"
                    )
                    num_passes_b = gr.Textbox(
                        label="Number of rerank passes B", value="1"
                    )
            else:
                num_passes = gr.Textbox(label="Number of rerank passes", value="1")
            qid = gr.Number(label="QID", value=1)

    if side_by_side:
        return [
            dataset,
            top_k_retrieve,
            top_k_rerank,
            host_retriever_a,
            host_retriever_b,
            host_reranker_a,
            host_reranker_b,
            num_passes_a,
            num_passes_b,
            qid,
        ]
    else:
        return [
            dataset,
            top_k_retrieve,
            top_k_rerank,
            host_retriever,
            host_reranker,
            num_passes,
            qid,
        ]


with gr.Blocks() as demo:
    gr.HTML(tooltip_style)
    gr.HTML(html_content)

    with gr.Tab("⚔️ ragnarok (side-by-side unblinded)"):
        with gr.Row():
            retriever_a, reranker_a, llm_a = rag_pipeline_block(label_suffix="A")
            retriever_b, reranker_b, llm_b = rag_pipeline_block(label_suffix="B")

        input_text, button = input_block()
        pretty_output_a, pretty_output_b, json_output_a, json_output_b = output_block(
            side_by_side=True
        )
        (
            a_better_button,
            b_better_button,
            both_good_button,
            both_bad_button,
        ) = comparison_block()

        (
            dataset,
            top_k_retrieve,
            top_k_rerank,
            host_retriever_a,
            host_retriever_b,
            host_reranker_a,
            host_reranker_b,
            num_passes_a,
            num_passes_b,
            qid,
        ) = parameters_block(side_by_side=True)

        button.click(
            on_submit_side_by_side,
            inputs=[
                llm_a,
                llm_b,
                retriever_a,
                retriever_b,
                reranker_a,
                reranker_b,
                dataset,
                host_retriever_a,
                host_reranker_a,
                host_retriever_b,
                host_reranker_b,
                top_k_retrieve,
                top_k_rerank,
                qid,
                input_text,
                num_passes_a,
                num_passes_b,
            ],
            outputs=[pretty_output_a, pretty_output_b, json_output_a, json_output_b],
        )

    with gr.Tab("⚔️ ragnarok (side-by-side blinded)"):
        with gr.Row():
            gr.Textbox(
                value="Unknown retriever + reranker + generation model",
                label="System A",
            )
            gr.Textbox(
                value="Unknown retriever + reranker + generation model",
                label="System B",
            )

        input_text, button = input_block()

        pretty_output_a, pretty_output_b, json_output_a, json_output_b = output_block(
            side_by_side=True
        )
        (
            a_better_button,
            b_better_button,
            both_good_button,
            both_bad_button,
        ) = comparison_block()

        (
            dataset,
            top_k_retrieve,
            top_k_rerank,
            host_retriever_a,
            host_retriever_b,
            host_reranker_a,
            host_reranker_b,
            num_passes_a,
            num_passes_b,
            qid,
        ) = parameters_block(side_by_side=True)

        button.click(
            on_submit_side_by_side_blinded,
            inputs=[
                dataset,
                host_retriever_a,
                host_reranker_a,
                host_retriever_b,
                host_reranker_b,
                top_k_retrieve,
                top_k_rerank,
                qid,
                input_text,
            ],
            outputs=[pretty_output_a, pretty_output_b, json_output_a, json_output_b],
        )

    with gr.Tab("💬 Direct Chat"):
        retriever, reranker, llm = rag_pipeline_block()

        input_text, button = input_block_direct()
        pretty_output, json_output = output_block(side_by_side=False)

        (
            dataset,
            top_k_retrieve,
            top_k_rerank,
            host_retriever,
            host_reranker,
            num_passes,
            qid,
        ) = parameters_block(side_by_side=False)

        def on_submit_single(
            model,
            retriever,
            reranker,
            dataset,
            host_retriever,
            host_reranker,
            top_k_retrieve,
            top_k_rerank,
            qid,
            query,
        ):
            def query_wrapper(
                retriever, reranker, model, host_retriever, host_reranker
            ):
                return query_model(
                    retriever,
                    reranker,
                    model,
                    dataset,
                    host_retriever,
                    host_reranker,
                    top_k_retrieve,
                    top_k_rerank,
                    qid,
                    query,
                )

            result, response = query_wrapper(
                retriever, reranker, model, host_retriever, host_reranker
            )

            return [result, response]

        button.click(
            on_submit_single,
            inputs=[
                llm,
                retriever,
                reranker,
                dataset,
                host_retriever,
                host_reranker,
                top_k_retrieve,
                top_k_rerank,
                qid,
                input_text,
            ],
            outputs=[pretty_output, json_output],
        )

    with gr.Tab("🏆 Leaderboard"):
        html_content = """
        <div class='navbar'>
            <h2>Ragnarök Chatbot Arena Leaderboard</h2>
        </div>
        """
        gr.HTML(html_content)

demo.launch()



import json
import os
import urllib.request
from urllib.parse import urlparse

from pyserini.util import *


def no_bool_convert(pairs):
    return {k: str(v).casefold() if isinstance(v, bool) else v for k, v in pairs}


def parse_file_info(file_name, file_url, save_dir):
    # download file
    local_file_path = download_url(
        file_url, save_dir, local_filename=file_name, verbose=True
    )

    # Compute MD5
    md5 = compute_md5(local_file_path)

    # Get file size
    size = os.path.getsize(local_file_path)

    # Remove the downloaded file
    os.remove(local_file_path)

    # Placeholder description
    description = "Sample description"

    return description, md5, size


# Save directory
def get_repo_info(repo_url, save_dir):
    parsed_url = urlparse(repo_url)
    path_parts = parsed_url.path.strip("/").split("/")
    if len(path_parts) < 2:
        raise ValueError("Invalid GitHub repository URL")

    owner, repo = path_parts[:2]
    api_url = f"https://api.github.com/repos/{owner}/{repo}/git/trees/main?recursive=1"
    with urllib.request.urlopen(api_url) as response:
        data = json.loads(response.read().decode())

    hits_info = {}
    for file in data.get("tree", []):
        if file["type"] == "blob":
            file_name = file["path"]
            file_url = f"https://github.com/{owner}/{repo}/raw/main/{file_name}"
            description, md5, size = parse_file_info(file_name, file_url, save_dir)
            hits_info["/".join(file_name.rsplit("/", 2)[-2:])] = {
                "description": description,
                "urls": [file_url],
                "md5": md5,
                "size (bytes)": size,
                "downloaded": False,
            }
    return hits_info


repo_url = "https://github.com/castorini/rank_llm_data/tree/main/retrieve_results"
save_dir = "retrieve_results/SPLADE_P_P_ENSEMBLE_DISTIL"
info = get_repo_info(repo_url, save_dir)
info_json = json.dumps(info, indent=4)

with open("repo_info.txt", "w") as f:
    f.write(f"HITS_INFO = {info_json}\n")

with open("repo_info.txt", "r") as f:
    data = f.read()
    data = data.replace("false", "False")

with open("repo_info.py", "w") as f:
    f.write(data)

os.remove("repo_info.txt")



import time
from typing import Any, Dict, List, Tuple, Union

import cohere
from ftfy import fix_text

from ragnarok.data import RAGExecInfo, Request
from ragnarok.generate.api_keys import get_cohere_api_key
from ragnarok.generate.llm import LLM, PromptMode
from ragnarok.generate.post_processor import CoherePostProcessor


class Cohere(LLM):
    def __init__(
        self,
        model: str,
        context_size: int,
        prompt_mode: PromptMode = PromptMode.COHERE,
        max_output_tokens: int = 1500,
        num_few_shot_examples: int = 0,
        key: str = get_cohere_api_key(),
    ) -> None:
        """
        Creates instance of the Cohere class, to deal with Cohere Command R models.

        Parameters:
        - model (str): The model identifier for the LLM (model identifier information can be found via OpenAI's model lists).
        - context_size (int): The maximum number of tokens that the model can handle in a single request.
        - prompt_mode (PromptMode, optional): Specifies the mode of prompt generation, with the default set to RANK_GPT,
         indicating that this class is designed primarily for listwise ranking tasks following the RANK_GPT methodology.
        - max_output_tokens (int, optional): Maximum number of tokens that can be generated in a single response. Defaults to 1500.
        - num_few_shot_examples (int, optional): Number of few-shot learning examples to include in the prompt, allowing for
        the integration of example-based learning to improve model performance. Defaults to 0, indicating no few-shot examples
        by default.
        - key (str, optional): The Cohere API key, defaults to the value of the COHERE_API_KEY environment variable.

        Raises:
        - ValueError: If an unsupported prompt mode is provided or if no Cohere API key / invalid key is supplied.
        """
        if model not in ["command-r-plus", "command-r"]:
            raise ValueError(
                f"Unsupported model: {model}. The only models currently supported are 'command-r' and 'command-r-plus' in Cohere."
            )
        super().__init__(
            model, context_size, prompt_mode, max_output_tokens, num_few_shot_examples
        )
        self._client = cohere.Client(key)
        self._post_processor = CoherePostProcessor()

    def run_llm(
        self,
        prompt: Union[str, List[Dict[str, Any]]],
        logging: bool = False,
    ) -> Tuple[Any, RAGExecInfo]:
        query, top_k_docs = prompt[0]["query"], prompt[0]["context"]
        if logging:
            print(f"Query: {query}")
            print(f"Top K Docs: {top_k_docs}")
        while True:
            try:
                response = self._client.chat(
                    model=self._model, message=query, documents=top_k_docs
                )
                break
            except Exception as e:
                print(str(e))
                if "blocked output" in str(e):
                    answers = []
                    rag_exec_info = RAGExecInfo(
                        prompt=prompt[0],
                        response="Blocked output",
                        input_token_count=len(query.split())
                        + sum([len(doc["snippet"].split()) for doc in top_k_docs]),
                        output_token_count=0,
                        candidates=top_k_docs,
                    )
                    return answers, rag_exec_info
                time.sleep(60)
        if logging:
            print(f"Response: {response}")
        answers, rag_exec_response = self._post_processor(response)
        if logging:
            print(f"Answers: {answers}")
        rag_exec_info = RAGExecInfo(
            prompt=prompt[0],
            response=rag_exec_response,
            input_token_count=len(query.split())
            + sum([len(doc["snippet"].split()) for doc in top_k_docs]),
            output_token_count=sum([len(ans.text) for ans in answers]),
            candidates=top_k_docs,
        )

        return answers, rag_exec_info

    def create_prompt(
        self, request: Request, topk: int
    ) -> Tuple[List[Dict[str, Any]], int]:
        query = request.query.text
        max_length = (self._context_size - 200) // topk
        self._prompt_mode = PromptMode.COHERE
        while True:
            rank = 0
            context = []
            for cand in request.candidates[:topk]:
                rank += 1
                content = self.convert_doc_to_prompt_content(cand.doc, max_length)
                context.append(content)
            if self._prompt_mode == PromptMode.COHERE:
                messages = [{"query": query, "context": context}]
            num_tokens = self.get_num_tokens(messages)
            if num_tokens <= self.max_tokens() - self.num_output_tokens():
                break
            else:
                max_length -= max(
                    1,
                    (num_tokens - self.max_tokens() + self.num_output_tokens())
                    // (topk * 4),
                )
        return messages, self.get_num_tokens(messages)

    def get_num_tokens(self, prompt: Union[str, List[Dict[str, str]]]) -> int:
        """Returns the number of tokens used by a list of messages in prompt."""
        # TODO(ronak): Add support
        return -1

    def cost_per_1k_token(self, input_token: bool) -> float:
        # TODO(ronak): Add support
        return -1

    def convert_doc_to_prompt_content(
        self, doc: Dict[str, Any], max_length: int
    ) -> str:
        if "text" in doc:
            content = doc["text"]
        elif "segment" in doc:
            content = doc["segment"]
        elif "contents" in doc:
            content = doc["contents"]
        else:
            content = doc["passage"]
        content = {"snippet": content}
        if "title" in doc:
            content["title"] = doc["title"]
        for key in content:
            content[key] = content[key].strip()
            content[key] = fix_text(content[key])
            content[key] = content[key].replace("\n", " ")
            content[key] = " ".join(content[key].split()[: int(max_length)])
            content[key] = self._replace_number(content[key])
        return content
